@@ -1,6 +1,6 @@
 /*
  * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
- * Copyright (C) 1999-2007 Hiroyuki Yamamoto
+ * Copyright (C) 1999-2010 Hiroyuki Yamamoto
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -75,11 +75,17 @@
 #include "logwindow.h"
 #include "folder.h"
 #include "setup.h"
+#include "sylmain.h"
 #include "utils.h"
 #include "gtkutils.h"
 #include "socket.h"
 #include "stock_pixmap.h"
 #include "trayicon.h"
+#include "plugin.h"
+#include "plugin_manager.h"
+#include "foldersel.h"
+#include "update_check.h"
+#include "colorlabel.h"
 
 #if USE_GPGME
 #  include "rfc2015.h"
@@ -93,16 +99,25 @@
 #  include <windows.h>
 #  include <pbt.h>
 #  include <fcntl.h>
+#  include <conio.h>
 #endif
 
 #include "version.h"
 
 gchar *prog_version;
 
+#ifdef G_OS_WIN32
+static gboolean init_console_done = FALSE;
+#endif
+
 static gint lock_socket = -1;
 static gint lock_socket_tag = 0;
 static GIOChannel *lock_ch = NULL;
 
+#if USE_THREADS
+static GThread *main_thread;
+#endif
+
 static struct RemoteCmd {
 	gboolean receive;
 	gboolean receive_all;
@@ -122,6 +137,19 @@ static struct RemoteCmd {
 #endif
 } cmd;
 
+#define STATUSBAR_PUSH(mainwin, str) \
+{ \
+	gtk_statusbar_push(GTK_STATUSBAR(mainwin->statusbar), \
+			   mainwin->mainwin_cid, str); \
+	gtkut_widget_draw_now(mainwin->statusbar); \
+}
+
+#define STATUSBAR_POP(mainwin) \
+{ \
+	gtk_statusbar_pop(GTK_STATUSBAR(mainwin->statusbar), \
+			  mainwin->mainwin_cid); \
+}
+
 static void parse_cmd_opt		(int		 argc,
 					 char		*argv[]);
 
@@ -130,9 +158,8 @@ static void parse_gtkrc_files		(void);
 static void setup_rc_dir		(void);
 static void check_gpg			(void);
 static void set_log_handlers		(gboolean	 enable);
-#ifdef G_OS_WIN32
-static void process_win32_message	(void);
-#endif
+static void register_system_events	(void);
+static void plugin_init			(void);
 
 static gchar *get_socket_name		(void);
 static gint prohibit_duplicate_launch	(void);
@@ -171,22 +198,24 @@ static void send_queue			(void);
 		exit(val);		\
 }
 
+static void load_cb(GObject *obj, GModule *module, gpointer data)
+{
+	g_print("load_cb: %p (%s), %p\n", module, module ? g_module_name(module) : "(null)", data);
+}
 
 int main(int argc, char *argv[])
 {
 	MainWindow *mainwin;
 	FolderView *folderview;
 	GdkPixbuf *icon;
+#ifdef G_OS_WIN32
 	GList *iconlist = NULL;
+#endif
+	GObject *syl_app;
 
 	app_init();
 	parse_cmd_opt(argc, argv);
 
-	sock_init();
-#if USE_SSL
-	ssl_init();
-#endif
-
 	/* check and create (unix domain) socket for remote operation */
 	lock_socket = prohibit_duplicate_launch();
 	if (lock_socket < 0) return 0;
@@ -197,20 +226,18 @@ int main(int argc, char *argv[])
 		return 0;
 	}
 
+#if USE_THREADS
+	gdk_threads_enter();
+#endif
 	gtk_set_locale();
 	gtk_init(&argc, &argv);
 
+	syl_app = syl_app_create();
+
 	gdk_rgb_init();
 	gtk_widget_set_default_colormap(gdk_rgb_get_cmap());
 	gtk_widget_set_default_visual(gdk_rgb_get_visual());
 
-#if USE_THREADS || USE_LDAP
-	if (!g_thread_supported())
-		g_thread_init(NULL);
-	if (!g_thread_supported())
-		g_error(_("g_thread is not supported by glib.\n"));
-#endif
-
 	parse_gtkrc_files();
 	setup_rc_dir();
 
@@ -230,9 +257,11 @@ int main(int argc, char *argv[])
 	CHDIR_EXIT_IF_FAIL(get_home_dir(), 1);
 
 	prefs_common_read_config();
+	filter_set_addressbook_func(addressbook_has_address);
 	filter_read_config();
 	prefs_actions_read_config();
 	prefs_display_header_read_config();
+	colorlabel_read_config();
 
 #ifdef G_OS_WIN32
 	{
@@ -296,34 +325,46 @@ int main(int argc, char *argv[])
 
 	set_log_handlers(TRUE);
 
-#ifdef G_OS_WIN32
-	process_win32_message();
-#endif
-
 	account_read_config_all();
 	account_set_menu();
 	main_window_reflect_prefs_all();
 
 	if (folder_read_list() < 0) {
-		setup(mainwin);
+		setup_mailbox();
 		folder_write_list();
 	}
 	if (!account_get_list()) {
-		account_edit_open();
-		account_add();
+		setup_account();
 	}
 
+	account_set_menu();
+	main_window_reflect_prefs_all();
+
 	account_set_missing_folder();
 	folder_set_missing_folders();
 	folderview_set(folderview);
 
 	addressbook_read_file();
 
+	register_system_events();
+
 	inc_autocheck_timer_init(mainwin);
 
+	plugin_init();
+
+	g_signal_emit_by_name(syl_app, "init-done");
+
 	remote_command_exec();
 
+#if USE_UPDATE_CHECK
+	if (prefs_common.auto_update_check)
+		update_check(FALSE);
+#endif
+
 	gtk_main();
+#if USE_THREADS
+	gdk_threads_leave();
+#endif
 
 	return 0;
 }
@@ -333,7 +374,6 @@ static void init_console(void)
 #ifdef G_OS_WIN32
 	gint fd;
 	FILE *fp;
-	static gboolean init_console_done = FALSE;
 
 	if (init_console_done)
 		return;
@@ -432,7 +472,7 @@ static void parse_cmd_opt(int argc, char *argv[])
 			while (p && *p != '\0' && *p != '-') {
 				if (!cmd.attach_files)
 					cmd.attach_files = g_ptr_array_new();
-				if (*p != G_DIR_SEPARATOR)
+				if (!g_path_is_absolute(p))
 					file = g_strconcat(get_startup_dir(),
 							   G_DIR_SEPARATOR_S,
 							   p, NULL);
@@ -564,74 +604,62 @@ static gint get_queued_message_num(void)
 	return queue->total;
 }
 
-static void app_init(void)
+#if USE_THREADS
+/* enables recursive locking with gdk_thread_enter / gdk_threads_leave */
+static GStaticRecMutex syl_mutex = G_STATIC_REC_MUTEX_INIT;
+
+static void thread_enter_func(void)
 {
-#ifdef G_OS_WIN32
-	gchar *newpath;
-	const gchar *lang_env;
-
-	/* disable locale variable such as "LANG=1041" */
-
-#define DISABLE_DIGIT_LOCALE(envstr)			\
-{							\
-	lang_env = g_getenv(envstr);			\
-	if (lang_env && g_ascii_isdigit(lang_env[0]))	\
-		g_unsetenv(envstr);			\
+	g_static_rec_mutex_lock(&syl_mutex);
+#if 0
+	syl_mutex_lock_count++;
+	if (syl_mutex_lock_count > 1)
+		g_print("enter: syl_mutex_lock_count: %d\n", syl_mutex_lock_count);
+#endif
 }
 
-	DISABLE_DIGIT_LOCALE("LC_ALL");
-	DISABLE_DIGIT_LOCALE("LANG");
-	DISABLE_DIGIT_LOCALE("LC_CTYPE");
-	DISABLE_DIGIT_LOCALE("LC_MESSAGES");
-
-#undef DISABLE_DIGIT_LOCALE
-
-	g_unsetenv("LANGUAGE");
+static void thread_leave_func(void)
+{
+#if 0
+	syl_mutex_lock_count--;
+	if (syl_mutex_lock_count > 0)
+		g_print("leave: syl_mutex_lock_count: %d\n", syl_mutex_lock_count);
 #endif
+	g_static_rec_mutex_unlock(&syl_mutex);
+}
 
-	setlocale(LC_ALL, "");
-
-	prog_version = PROG_VERSION;
-	set_startup_dir();
-
-#ifdef G_OS_WIN32
-	/* include startup directory into %PATH% for GSpawn */
-	newpath = g_strconcat(get_startup_dir(), ";", g_getenv("PATH"), NULL);
-	g_setenv("PATH", newpath, TRUE);
-	g_free(newpath);
-#endif
-
-	if (g_path_is_absolute(LOCALEDIR))
-		bindtextdomain(PACKAGE, LOCALEDIR);
-	else {
-		gchar *locale_dir;
-
-		locale_dir = g_strconcat(get_startup_dir(), G_DIR_SEPARATOR_S,
-					 LOCALEDIR, NULL);
-#ifdef G_OS_WIN32
-		{
-			gchar *locale_dir_;
-
-			locale_dir_ = g_locale_from_utf8(locale_dir, -1,
-							 NULL, NULL, NULL);
-			if (locale_dir_) {
-				g_free(locale_dir);
-				locale_dir = locale_dir_;
-			}
-		}
-#endif
-		bindtextdomain(PACKAGE, locale_dir);
-		g_free(locale_dir);
+static void event_loop_iteration_func(void)
+{
+	if (g_thread_self() != main_thread) {
+		g_fprintf(stderr, "event_loop_iteration_func called from non-main thread (%p)\n", g_thread_self());
+		g_usleep(10000);
+		return;
 	}
-
-	bind_textdomain_codeset(PACKAGE, CS_UTF_8);
-	textdomain(PACKAGE);
-
-#ifdef G_OS_UNIX
-	/* ignore SIGPIPE signal for preventing sudden death of program */
-	signal(SIGPIPE, SIG_IGN);
+	gtk_main_iteration();
+}
 #endif
 
+static void app_init(void)
+{
+#if USE_THREADS
+	if (!g_thread_supported())
+		g_thread_init(NULL);
+	if (!g_thread_supported())
+		g_error("g_thread is not supported by glib.");
+	else {
+		gdk_threads_set_lock_functions(thread_enter_func,
+					       thread_leave_func);
+		gdk_threads_init();
+		main_thread = g_thread_self();
+	}
+#endif
+	syl_init();
+
+#if USE_THREADS
+	set_event_loop_func(event_loop_iteration_func);
+#endif
+	prog_version = PROG_VERSION;
+
 #ifdef G_OS_WIN32
 	read_ini_file();
 #endif
@@ -703,24 +731,7 @@ static void setup_rc_dir(void)
 	}
 #endif /* !G_OS_WIN32 */
 
-	if (!is_dir_exist(get_rc_dir())) {
-		if (make_dir_hier(get_rc_dir()) < 0)
-			exit(1);
-	}
-
-	MAKE_DIR_IF_NOT_EXIST(get_mail_base_dir());
-
-	CHDIR_EXIT_IF_FAIL(get_rc_dir(), 1);
-
-	MAKE_DIR_IF_NOT_EXIST(get_imap_cache_dir());
-	MAKE_DIR_IF_NOT_EXIST(get_news_cache_dir());
-	MAKE_DIR_IF_NOT_EXIST(get_mime_tmp_dir());
-	MAKE_DIR_IF_NOT_EXIST(get_tmp_dir());
-	MAKE_DIR_IF_NOT_EXIST(UIDL_DIR);
-
-	/* remove temporary files */
-	remove_all_files(get_tmp_dir());
-	remove_all_files(get_mime_tmp_dir());
+	syl_setup_rc_dir();
 }
 
 void app_will_exit(gboolean force)
@@ -759,6 +770,8 @@ void app_will_exit(gboolean force)
 		manage_window_focus_in(mainwin->window, NULL, NULL);
 	}
 
+	g_signal_emit_by_name(syl_app_get(), "app-exit");
+
 	inc_autocheck_timer_remove();
 
 	if (prefs_common.clean_on_exit)
@@ -772,36 +785,26 @@ void app_will_exit(gboolean force)
 			procmsg_remove_all_cached_messages(FOLDER(ac->folder));
 	}
 
+	syl_plugin_unload_all();
+
 	trayicon_destroy(mainwin->tray_icon);
 
 	/* save all state before exiting */
-	folder_write_list();
 	summary_write_cache(mainwin->summaryview);
-
 	main_window_get_size(mainwin);
 	main_window_get_position(mainwin);
-	prefs_common_write_config();
-	filter_write_config();
-	account_write_config_all();
+	syl_save_all_state();
 	addressbook_export_to_file();
 
 	filename = g_strconcat(get_rc_dir(), G_DIR_SEPARATOR_S, MENU_RC, NULL);
 	gtk_accel_map_save(filename);
 	g_free(filename);
 
-	/* remove temporary files */
-	remove_all_files(get_tmp_dir());
-	remove_all_files(get_mime_tmp_dir());
-
-	set_log_handlers(FALSE);
-	close_log_file();
+	/* remove temporary files, close log file, socket cleanup */
+	syl_cleanup();
 	lock_socket_remove();
 
-#if USE_SSL
-	ssl_done();
-#endif
 	cleanup_console();
-	sock_cleanup();
 
 	if (gtk_main_level() > 0)
 		gtk_main_quit();
@@ -822,18 +825,31 @@ static void idle_function_for_gpgme(void)
 static void check_gpg(void)
 {
 #if USE_GPGME
-	if (gpgme_check_version("0.4.5") &&
-	    !gpgme_engine_check_version(GPGME_PROTOCOL_OpenPGP)) {
+	const gchar *version;
+	gpgme_error_t err = 0;
+
+	version = gpgme_check_version("1.0.0");
+	if (version) {
+		debug_print("GPGME Version: %s\n", version);
+		err = gpgme_engine_check_version(GPGME_PROTOCOL_OpenPGP);
+		if (err)
+			debug_print("gpgme_engine_check_version: %s\n",
+				    gpgme_strerror(err));
+	}
+
+	if (version && !err) {
 		/* Also does some gpgme init */
 	        gpgme_engine_info_t engineInfo;
 
+#if HAVE_LOCALE_H
 		gpgme_set_locale(NULL, LC_CTYPE, setlocale(LC_CTYPE, NULL));
 		gpgme_set_locale(NULL, LC_MESSAGES,
 				 setlocale(LC_MESSAGES, NULL));
+#endif
 
 		if (!gpgme_get_engine_info(&engineInfo)) {
 			while (engineInfo) {
-				debug_print("GpgME Protocol: %s\n      Version: %s\n",
+				debug_print("GPGME Protocol: %s\n      Version: %s\n",
 					    gpgme_get_protocol_name
 						(engineInfo->protocol),
 					    engineInfo->version ?
@@ -844,6 +860,7 @@ static void check_gpg(void)
 
 		procmsg_set_decrypt_message_func
 			(rfc2015_open_message_decrypted);
+		procmsg_set_auto_decrypt_message(TRUE);
 	} else {
 		rfc2015_disable_all();
 
@@ -964,10 +981,20 @@ static void set_log_handlers(gboolean enable)
 }
 
 #ifdef G_OS_WIN32
+static BOOL WINAPI
+ctrl_handler(DWORD dwctrltype)
+{
+	log_print("ctrl_handler: received %d\n", dwctrltype);
+	app_will_exit(TRUE);
+
+	return TRUE;
+}
+
 static LRESULT CALLBACK
 wndproc(HWND hwnd, UINT message, WPARAM wparam, LPARAM lparam)
 {
-	if (message == WM_POWERBROADCAST) {
+	switch (message) {
+	case WM_POWERBROADCAST:
 		debug_print("WM_POWERBROADCAST received: wparam = %d\n",
 			    wparam);
 		if (wparam == PBT_APMSUSPEND || wparam == PBT_APMSTANDBY) {
@@ -978,21 +1005,40 @@ wndproc(HWND hwnd, UINT message, WPARAM wparam, LPARAM lparam)
 			debug_print("resume now\n");
 			inc_autocheck_timer_set();
 		}
+		break;
+	case WM_ENDSESSION:
+		if (wparam == 1) {
+			log_print("WM_ENDSESSION received: system is quitting\n");
+			app_will_exit(TRUE);
+		}
+		break;
+	default:
+		break;
 	}
 
 	return DefWindowProc(hwnd, message, wparam, lparam);
 }
 
-static void process_win32_message(void)
+static void register_system_events(void)
 {
 	WNDCLASS wclass;
 	static HWND hwnd = NULL;
+	static BOOL ctrl_handler_set = FALSE;
 	ATOM klass;
 	HINSTANCE hmodule = GetModuleHandle(NULL);
 
+	if (init_console_done && !ctrl_handler_set) {
+		debug_print("register_system_events(): SetConsoleCtrlHandler\n");
+		ctrl_handler_set = SetConsoleCtrlHandler(ctrl_handler, TRUE);
+		if (!ctrl_handler_set)
+			g_warning("SetConsoleCtrlHandler() failed\n");
+	}
+
 	if (hwnd)
 		return;
 
+	debug_print("register_system_events(): RegisterClass\n");
+
 	memset(&wclass, 0, sizeof(WNDCLASS));
 	wclass.lpszClassName = "sylpheed-observer";
 	wclass.lpfnWndProc   = wndproc;
@@ -1007,8 +1053,135 @@ static void process_win32_message(void)
 	if (!hwnd)
 		UnregisterClass(MAKEINTRESOURCE(klass), hmodule);
 }
+#else /* G_OS_WIN32 */
+static void sig_handler(gint signum)
+{
+	debug_print("signal %d received\n", signum);
+
+	switch (signum) {
+	case SIGHUP:
+	case SIGINT:
+	case SIGTERM:
+	case SIGQUIT:
+		app_will_exit(TRUE);
+		break;
+	default:
+		break;
+	}
+}
+
+static void register_system_events(void)
+{
+	struct sigaction sa;
+
+	memset(&sa, 0, sizeof(sa));
+	sa.sa_handler = sig_handler;
+	sa.sa_flags = SA_RESTART;
+
+	sigemptyset(&sa.sa_mask);
+	sigaddset(&sa.sa_mask, SIGHUP);
+	sigaddset(&sa.sa_mask, SIGINT);
+	sigaddset(&sa.sa_mask, SIGTERM);
+	sigaddset(&sa.sa_mask, SIGQUIT);
+	sigaddset(&sa.sa_mask, SIGPIPE);
+
+	sigaction(SIGHUP, &sa, NULL);
+	sigaction(SIGINT, &sa, NULL);
+	sigaction(SIGTERM, &sa, NULL);
+	sigaction(SIGQUIT, &sa, NULL);
+	sigaction(SIGPIPE, &sa, NULL);
+}
 #endif
 
+static void plugin_init(void)
+{
+	MainWindow *mainwin;
+
+	mainwin = main_window_get();
+
+	STATUSBAR_PUSH(mainwin, _("Loading plug-ins..."));
+
+	if (syl_plugin_init_lib() != 0) {
+		STATUSBAR_POP(mainwin);
+		return;
+	}
+
+	syl_plugin_add_symbol("prog_version", prog_version);
+	syl_plugin_add_symbol("main_window_lock", main_window_lock);
+	syl_plugin_add_symbol("main_window_unlock", main_window_unlock);
+	syl_plugin_add_symbol("main_window_get", main_window_get);
+	syl_plugin_add_symbol("main_window_popup", main_window_popup);
+	syl_plugin_add_symbol("app_will_exit", app_will_exit);
+	syl_plugin_add_symbol("main_window_menu_factory",
+			      mainwin->menu_factory);
+	syl_plugin_add_symbol("main_window_statusbar", mainwin->statusbar);
+
+	syl_plugin_add_symbol("folderview", mainwin->folderview);
+	syl_plugin_add_symbol("folderview_get_selected_item",
+			      folderview_get_selected_item);
+	syl_plugin_add_symbol("folderview_mail_popup_factory",
+			      mainwin->folderview->mail_factory);
+	syl_plugin_add_symbol("folderview_imap_popup_factory",
+			      mainwin->folderview->imap_factory);
+	syl_plugin_add_symbol("folderview_news_popup_factory",
+			      mainwin->folderview->news_factory);
+
+	syl_plugin_add_symbol("summaryview", mainwin->summaryview);
+	syl_plugin_add_symbol("summary_select_by_msgnum",
+			      summary_select_by_msgnum);
+	syl_plugin_add_symbol("summary_select_by_msginfo",
+			      summary_select_by_msginfo);
+	syl_plugin_add_symbol("summary_lock", summary_lock);
+	syl_plugin_add_symbol("summary_unlock", summary_unlock);
+	syl_plugin_add_symbol("summary_is_locked", summary_is_locked);
+
+	syl_plugin_add_symbol("messageview_create_with_new_window",
+			      messageview_create_with_new_window);
+	syl_plugin_add_symbol("messageview_show", messageview_show);
+
+	syl_plugin_add_symbol("compose_new", compose_new);
+	syl_plugin_add_symbol("compose_entry_set", compose_entry_set);
+	syl_plugin_add_symbol("compose_entry_append", compose_entry_append);
+	syl_plugin_add_symbol("compose_entry_get_text", compose_entry_get_text);
+	syl_plugin_add_symbol("compose_lock", compose_lock);
+	syl_plugin_add_symbol("compose_unlock", compose_unlock);
+
+	syl_plugin_add_symbol("foldersel_folder_sel",
+			      foldersel_folder_sel);
+	syl_plugin_add_symbol("foldersel_folder_sel_full",
+			      foldersel_folder_sel_full);
+	syl_plugin_add_symbol("input_dialog", input_dialog);
+	syl_plugin_add_symbol("input_dialog_with_invisible",
+			      input_dialog_with_invisible);
+
+	syl_plugin_add_symbol("manage_window_set_transient",
+			      manage_window_set_transient);
+	syl_plugin_add_symbol("manage_window_signals_connect",
+			      manage_window_signals_connect);
+	syl_plugin_add_symbol("manage_window_get_focus_window",
+			      manage_window_get_focus_window);
+
+	syl_plugin_add_symbol("inc_mail", inc_mail);
+	syl_plugin_add_symbol("inc_is_active", inc_is_active);
+	syl_plugin_add_symbol("inc_lock", inc_lock);
+	syl_plugin_add_symbol("inc_unlock", inc_unlock);
+
+	syl_plugin_signal_connect("plugin-load", G_CALLBACK(load_cb), NULL);
+
+#ifdef G_OS_WIN32
+	{
+		gchar *path;
+		path = g_strconcat(get_startup_dir(), G_DIR_SEPARATOR_S,
+				   "plugins", NULL);
+		syl_plugin_load_all(path);
+		g_free(path);
+	}
+#else
+	syl_plugin_load_all(PLUGINDIR);
+#endif
+	STATUSBAR_POP(mainwin);
+}
+
 static gchar *get_socket_name(void)
 {
 	static gchar *filename = NULL;
@@ -1199,6 +1372,10 @@ static gboolean lock_socket_input_cb(GIOChannel *source, GIOCondition condition,
 	gint fd, sock;
 	gchar buf[BUFFSIZE];
 
+#if USE_THREADS
+	gdk_threads_enter();
+#endif
+
 	fd = g_io_channel_unix_get_fd(source);
 	sock = fd_accept(fd);
 	fd_gets(sock, buf, sizeof(buf));
@@ -1258,6 +1435,9 @@ static gboolean lock_socket_input_cb(GIOChannel *source, GIOCondition condition,
 		strretchomp(buf);
 		if (strlen(buf) < 6 || buf[4] != ' ') {
 			fd_close(sock);
+#if USE_THREADS
+			gdk_threads_leave();
+#endif
 			return TRUE;
 		}
 		open_message(buf + 5);
@@ -1268,6 +1448,10 @@ static gboolean lock_socket_input_cb(GIOChannel *source, GIOCondition condition,
 
 	fd_close(sock);
 
+#if USE_THREADS
+	gdk_threads_leave();
+#endif
+
 	return TRUE;
 }
 
